#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/malloc.h>
/* Note: CONFIG_KMOD is defined else TCFS modules did not work! */
#include <linux/kmod.h>
#include <linux/tcfs.h>
#include <linux/tcfs/tcfsapi.h>

static struct tcfs_cipher_hook *tcfs_hook_start=NULL;
static int loaded_successfully=0;

void tcfs_register_cipher (char *cipher, struct tcfs_cipher_operations *cipher_ops)
{
	struct tcfs_cipher_hook *tmp=tcfs_hook_start;

	if (!cipher_ops->init_key || !cipher_ops->cleanup_key
		|| !cipher_ops->encrypt || !cipher_ops->decrypt) {
		return;
	}

	if (tcfs_hook_start==NULL) {
		tcfs_hook_start=(struct tcfs_cipher_hook *)kmalloc (sizeof (struct tcfs_cipher_hook), GFP_KERNEL);
		strncpy (tcfs_hook_start->cipher_name, cipher,
						TCFS_ENGINE_FIELD_SIZE);
		tcfs_hook_start->cipher_ops=cipher_ops;
		tcfs_hook_start->next=NULL;

		loaded_successfully=1;

		return; 
	}

	do {
		if (!memcmp (cipher, tmp->cipher_name,
				 TCFS_ENGINE_FIELD_SIZE)) {
			loaded_successfully=1;

			return ;
		}       
	} while (tmp->next && (tmp=tmp->next));

	tmp->next=(struct tcfs_cipher_hook *)kmalloc (sizeof (struct tcfs_cipher_hook), GFP_KERNEL);
	tmp=tmp->next;
	memcpy (tmp->cipher_name, cipher, TCFS_ENGINE_FIELD_SIZE);
	tmp->cipher_ops=cipher_ops;
	tmp->next=NULL;
	loaded_successfully=1;
}

void tcfs_unregister_cipher (char *cipher)
{
	struct tcfs_cipher_hook *tmp=tcfs_hook_start;
	struct tcfs_cipher_hook *prev=tcfs_hook_start;

	do {
		if (tmp && !memcmp(tmp->cipher_name, cipher,
					TCFS_ENGINE_FIELD_SIZE)) {
			prev->next=tmp->next;

			if (tmp==tcfs_hook_start) {
				tcfs_hook_start=NULL;
			}

			kfree (tmp);

			return;
		}
		prev=tmp;
	} while (tmp && (tmp=tmp->next));
}

struct tcfs_cipher_operations *tcfs_cipher_lookup (char *cipher_name)
{
	struct tcfs_cipher_hook *tmp=tcfs_hook_start;
	char *cipher_module=NULL;
	char *default_cipher=TCFS_DEFAULT_CIPHER;
	int res=0;

	if (!cipher_name) {
		printk ("cipher_lookup called with cipher_name=NULL!\n");
		cipher_name=default_cipher;
	}

	do {
		if (tmp && (!memcmp(tmp->cipher_name, cipher_name,
						TCFS_ENGINE_FIELD_SIZE))) {
				return tmp->cipher_ops;
		}
	} while (tmp && (tmp=tmp->next));

	cipher_module=(char *)kmalloc (13+strlen(cipher_name), GFP_KERNEL);
	if (!cipher_module) {
		printk ("TCFS: not enough memory\n");
		return NULL;
	}

	sprintf (cipher_module, "tcfs_%s_cipher", cipher_name);

	loaded_successfully=0;
	res=request_module (cipher_module);
	kfree (cipher_module);

	if (res) {
		printk ("TCFS: an error occurred loading tcfs_%s_cipher.\n",
			cipher_name);
		return NULL;
	}

	if (loaded_successfully) {
		return tcfs_cipher_lookup (cipher_name);
	}

	return NULL;
}

int tcfs_encrypt (char *block, int bytes, void *key, char *cipher)
{
	struct tcfs_cipher_operations *tmp=NULL;

	if (!cipher) {
		printk ("tcfs_encrypt called w/out a cipher!\n");
	}

	tmp=tcfs_cipher_lookup (cipher);

	if (tmp) {
		tmp->encrypt(block, bytes, key);
	} else {
		return -TCFS_NOCIPHER;
	}

	return -TCFS_ERROK;
}

int tcfs_decrypt (char *block, int bytes, void *key, char *cipher)
{
	struct tcfs_cipher_operations *tmp=NULL;

	if (!cipher) {
		printk ("tcfs_decrypt called w/out a cipher!\n");
	}

	tmp=tcfs_cipher_lookup (cipher);

	if (tmp) {
		tmp->decrypt(block, bytes, key);
	} else {
		return -TCFS_NOCIPHER;
	}

	return -TCFS_ERROK;
}

void *tcfs_init_key (char *key, char *cipher)
{
	struct tcfs_cipher_operations *tmp=NULL;

	if (strlen(cipher)==0) {
		cipher=NULL;
	}

	if (!cipher) {
		printk ("tcfs_init_key called w/out a cipher!\n");
	}

	tmp=tcfs_cipher_lookup (cipher);

	if (tmp) {
		return tmp->init_key(key);
	} else {
		return NULL;
	}
}

int tcfs_cleanup_key (void *key, char *cipher)
{
	struct tcfs_cipher_operations *tmp=NULL;

	if (!key) {
		return -TCFS_ERROK;
	}

	tmp=tcfs_cipher_lookup (cipher);

	if (tmp) {
		tmp->cleanup_key (key);
	} else {
		return -TCFS_NOCIPHER;
	}

	return -TCFS_ERROK;
}
