#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/semaphore.h>

#include <linux/tcfs/tcfs_hash.h>
#include <linux/tcfs/cipherinterface.h>

extern void tcfs_interpolate(struct user_shares *,unsigned int,unsigned char *);

struct user_key *user_key_table[HASH_SIZE];
struct group_key *group_key_table[HASH_SIZE];

struct semaphore tcfs_user_sem = MUTEX;
struct semaphore tcfs_group_sem = MUTEX;
   
#define tcfs_hash(id)	(id%HASH_SIZE)

void 
tcfs_init_hash (void)
{
	int i;

	for (i=0;i<HASH_SIZE;i++) {
		user_key_table[i]=NULL;
		group_key_table[i]=NULL;
	}
}

void 
tcfs_cleanup_hash(void)
{
	int i;
	struct user_key *tmp, *tmp1;
	struct group_key *tmp2, *tmp3;

	down(&tcfs_user_sem);

	for (i=0;i<HASH_SIZE;i++) {
		tmp=user_key_table[i];

		while (tmp!=NULL) {
			tmp1=tmp->next;
			if (tmp->key) {
				kfree (tmp->key);
			}

			kfree(tmp);
			tmp=tmp1;
		}
	}

	up(&tcfs_user_sem);
	down(&tcfs_group_sem);

	for (i=0;i<HASH_SIZE;i++) {
		tmp2=group_key_table[i];

		while (tmp2!=NULL) {
			struct user_shares *shtmp=NULL, *shtmp1;

			tmp3=tmp2->next;
			if (tmp2->key) {
				kfree (tmp2->key);
			}
			
			while (shtmp!=NULL) {
				shtmp1=shtmp->next;
				if (shtmp->share) {
					kfree (shtmp->share);
				}
				kfree (shtmp);
				shtmp=shtmp1;
			}

			kfree (tmp2);
			tmp2=tmp3;
		}
	}

	up(&tcfs_group_sem);
}

struct user_key * 
tcfs_key_lookup(__kernel_uid_t uid, char *engine)
{
	struct user_key *res=NULL;

	down(&tcfs_user_sem);

	res=user_key_table[tcfs_hash (uid)];

	while (res!=NULL && (res->uid!=uid || memcmp(res->engine, engine,
						TCFS_ENGINE_FIELD_SIZE))) {
		res=res->next;
	}


	if (res && (res->uid!=uid || memcmp(res->engine, engine,
						TCFS_ENGINE_FIELD_SIZE))) {
		res=NULL;
	}

	up(&tcfs_user_sem);

	return res;
}

int
tcfs_add_userkey (struct tcfs_user_key *uk, unsigned int flags)
{
	struct user_key *tmp;

	tmp=tcfs_key_lookup (uk->uid, uk->engine);

	down (&tcfs_user_sem);

	if (tmp) {
		if (!flags) {
			tmp->flags|=flags;
		} else {
			tmp->count++;
		} 
	} else {
		unsigned int huid;

		huid=tcfs_hash(uk->uid);
		tmp=user_key_table[huid];

		if (tmp==NULL) {
			user_key_table[huid]=(struct user_key *)
				kmalloc (sizeof(struct user_key), GFP_KERNEL);
			tmp=user_key_table[huid];
		} else {
			while (tmp->next != NULL) {
				tmp = tmp->next;
			};
		
			tmp->next=(struct user_key *) kmalloc (
					sizeof(struct user_key), GFP_KERNEL);
			if (!tmp->next) {
				up (&tcfs_user_sem);
	
				return -TCFS_ERRGEN;
			}
	
			tmp=tmp->next;
		}

		tmp->uid=uk->uid;
		memcpy (tmp->engine, uk->engine, TCFS_ENGINE_FIELD_SIZE);
		tmp->count=1;
		tmp->flags=0;
		tmp->key=tcfs_init_key (uk->key, uk->engine);
		tmp->next=NULL;
	}

	up (&tcfs_user_sem);

	return -TCFS_ERROK;
}

int
tcfs_rem_userkey (struct tcfs_user_key *uk, unsigned int flags)
{
	struct user_key *tmp;

	tmp=tcfs_key_lookup (uk->uid, uk->engine);

	if (!tmp) {
		return -TCFS_ERROK;
	} else {
		down (&tcfs_user_sem);

		if (flags&=TCFS_ALLKEY_FLAG) {
			tmp->count=0;
		} else {
			if (tmp->count > 0) {
				tmp->count--;
			}
		}

		if (flags&=TCFS_PERMANENT_FLAG) {
			tmp->flags&=~TCFS_PERMANENT_FLAG;
		}

		if (!tmp->count && !TCFS_IS_PERMANENT(tmp->flags)) {
			struct user_key *tmp1;

			tmp1=user_key_table[tcfs_hash(uk->uid)];
			if (tmp==tmp1) {
				user_key_table[tcfs_hash(uk->uid)]=tmp->next;
			} else {
				while (tmp1->next!=tmp) {
					tmp1=tmp1->next;
				};

				tmp1->next=tmp->next;
			}

			tcfs_cleanup_key (tmp->key, uk->engine);
			kfree (tmp);
		}
	}

	up (&tcfs_user_sem);

	return -TCFS_ERROK;
}

struct group_key *
tcfs_groupkey_lookup (__kernel_gid_t gid, char *engine)
{
	struct group_key *res=NULL;

	down (&tcfs_group_sem);

	res=group_key_table[tcfs_hash(gid)];

	while (res!=NULL && (res->gid!=gid && memcmp(res->engine, engine,
					TCFS_ENGINE_FIELD_SIZE))) {
		res=res->next;
	}

	up (&tcfs_group_sem);

	return res;
}

int
tcfs_add_groupshare (struct tcfs_group_share *gs, unsigned int flags)
{
	struct group_key *tmp=NULL;
	struct user_shares *stmp=NULL;

	tmp=tcfs_groupkey_lookup (gs->gid, gs->engine);
	if (tmp) {
		stmp=tmp->shares;
		while (stmp!=NULL && stmp->uid!=gs->uid) {
			stmp=stmp->next;
		};

		if (stmp) {
			if (!flags) {
				stmp->flags|=flags;
			} else {
				stmp->count++;
			}

			up (&tcfs_group_sem);

			return -TCFS_ERROK;
		} else {
			tmp->num_shares++;
		}

		stmp=tmp->shares;
		while (stmp->next != NULL) {
			stmp=stmp->next;
		};

		stmp->next=(struct user_shares *)kmalloc (
			sizeof (struct user_shares), GFP_KERNEL);
		stmp=stmp->next;
	} else {
		tmp=(struct group_key *)kmalloc (sizeof (struct group_key),
				GFP_KERNEL);
		tmp->gid=gs->gid;
		memcpy (tmp->engine, gs->engine, TCFS_ENGINE_FIELD_SIZE);
		tmp->num_shares=1;
		tmp->key=NULL;
		tmp->next=NULL;
		tmp->shares=(struct user_shares *)kmalloc (
			sizeof (struct user_shares), GFP_KERNEL);
		stmp=tmp->shares;
	}

	stmp->uid=gs->uid;
	stmp->count=1;
	stmp->flags=flags;
	stmp->share=(unsigned char *)kmalloc (GROUP_KEYSIZE, GFP_KERNEL);
	memcpy (stmp->share, gs->share, GROUP_KEYSIZE);
	stmp->next=NULL;

	if (tmp->num_shares==gs->threshold) {
		unsigned char *key;

		key=(char *)kmalloc (KEYSIZE, GFP_KERNEL);
		memset (key, 0, KEYSIZE);
		tcfs_interpolate (tmp->shares, gs->threshold, key);
		tmp->key=tcfs_init_key (key, gs->engine);
	}

	return -TCFS_ERROK;
}

int
tcfs_rem_groupshare (struct tcfs_group_share *gs, unsigned int flags)
{
	struct group_key *tmp, *tmp1;
	struct user_shares *stmp, *stmp1;

	tmp=tcfs_groupkey_lookup (gs->gid, gs->engine);

	if (!tmp) {
		return -TCFS_ERROK;
	} else {
		down (&tcfs_group_sem);
		stmp=tmp->shares;

		while (stmp!=NULL && stmp->uid!=gs->uid) {
			stmp=stmp->next;
		};

		if (!stmp) {
			up (&tcfs_group_sem);

			return -TCFS_ERROK;
		} else {
			if (flags&=TCFS_ALLKEY_FLAG) {
				stmp->count=0;
			} else {
				if (stmp->count > 0) {
					stmp->count--;
				}
			}

			if (flags&=TCFS_PERMANENT_FLAG) {
				stmp->flags&=~TCFS_PERMANENT_FLAG;
			}

			if (!stmp->count && !TCFS_IS_PERMANENT(stmp->flags)) {
				stmp1=tmp->shares;
				if (stmp1==stmp) {
					tmp->shares=stmp->next;
				} else {
					while (stmp1->next!=stmp) {
						stmp1=stmp1->next;
					};

					stmp1->next=stmp->next;
				}

				kfree (stmp->share);
				kfree (stmp);
			}
		}
	}

	tmp->num_shares--;
	if (!tmp->num_shares) {
		stmp=tmp->shares;
		stmp1=stmp->next;

		while (stmp!=NULL) {
			kfree (stmp->share);
			kfree (stmp);
			stmp=stmp1;
			stmp1=stmp->next;
		};

		tcfs_cleanup_key (tmp->key, gs->engine);
		tmp1=group_key_table[tcfs_hash(gs->gid)];
		while (tmp1!=NULL && tmp1->next!=tmp) {
			tmp1=tmp1->next;
		};

		tmp1->next=tmp->next;
		kfree (tmp);
	}

	up (&tcfs_group_sem);

	return -TCFS_ERROK;
}
